//
//  AllProcessesView.m
//  UpTop
//
//  Created by Geoff Prewett on 3/4/17.
//  Copyright © 2017 Geoff Prewett. All rights reserved.
//

#import "AllProcessesView.h"

#import "BarGraph.h"
#import "Utils.h"

typedef NSString* (*DataFunc)(Process *p);

@interface Column : NSTableColumn
@property NSTextAlignment alignment;
@property DataFunc dataFunc;

- (instancetype)initWithIdentifier:(NSString *)identifier;
@end

@implementation Column
- (instancetype)initWithIdentifier:(NSString *)identifier
{
    self = [super initWithIdentifier:identifier];
    if (self) {
    }
    return self;
}
@end

static NSString* getPID(Process *p)
{
    return [NSString stringWithFormat:@"%d", p.processID];
}

static NSString* getUserName(Process *p)
{
    return [NSString stringWithFormat:@"%@", p.userName];
}

static NSString* getLocalizedProcessName(Process *p)
{
    return [NSString stringWithFormat:@"%@", p.localizedProcessName];
}

static NSString* getProcessName(Process *p)
{
    return [NSString stringWithFormat:@"%@", p.processName];
}

static NSString* getCPU(Process *p)
{
    return [NSString stringWithFormat:@"%.1f%%", p.cpuPercent];
}

static NSString* getMemoryTotal(Process *p)
{
    return [NSString stringWithFormat:@"%.1f MB", toMB((uint64_t)p.totalMem)];
}

static NSString* getDiskReadRecent(Process *p)
{
    return [NSString stringWithFormat:@"%.1f MB", toMB(p.diskRead)];
}

static NSString* getDiskReadTotal(Process *p)
{
    return [NSString stringWithFormat:@"%.1f MB", toMB(p.totalDiskRead)];
}

static NSString* getDiskWrittenRecent(Process *p)
{
    return [NSString stringWithFormat:@"%.1f MB", toMB(p.diskWritten)];
}

static NSString* getDiskWrittenTotal(Process *p)
{
    return [NSString stringWithFormat:@"%.1f MB", toMB(p.totalDiskWritten)];
}

static Column* makeColumn(NSString *name, CGFloat width, NSTextAlignment align, NSString *field, DataFunc dataFunc)
{
    Column *c = [[Column alloc] initWithIdentifier:name];
    c.title = name;
    c.width = width;
    c.alignment = align;
    c.dataFunc = dataFunc;

    NSSortDescriptor *sorter = [NSSortDescriptor sortDescriptorWithKey:field
                                                             ascending:NO
                                                              selector:@selector(compare:)];
    [c setSortDescriptorPrototype:sorter];

    return c;
}

//-----------------------------------------------------------------------------
@interface AllProcessesView ()
@property CGFloat em;
@property NSArray *columns;
@property NSArray *processes;
@property NSScrollView *procTableContainer;
@property NSTableView *procTable;
@property NSTextField *memLabel;
@property BarGraph *memBar;
@end

@implementation AllProcessesView

- (instancetype)initWithFrame:(NSRect)rect
{
    self = [super initWithFrame:rect];
    if (self) {
        NSTextField *tmp = createLabel(@"Ag");
        [tmp sizeToFit];
        self.em = tmp.frame.size.height + 2;  // not sure why I need the +2

        self.columns = @[
                makeColumn(@"PID", 2.5 * self.em, NSTextAlignmentRight, @"processID", getPID),
                makeColumn(@"User", 4 * self.em, NSTextAlignmentCenter, @"userName", getUserName),
                makeColumn(@"Name", 10 * self.em, NSTextAlignmentLeft, @"localizedProcessName", getLocalizedProcessName),
                makeColumn(@"Executable", 10 * self.em, NSTextAlignmentLeft, @"processName", getProcessName),
                makeColumn(@"% CPU", 3 * self.em, NSTextAlignmentRight, @"cpuPercent", getCPU),
                makeColumn(@"Memory (total)", 5 * self.em, NSTextAlignmentRight, @"totalMem", getMemoryTotal),
                makeColumn(@"Disk read (recent)", 5 * self.em, NSTextAlignmentRight, @"diskRead", getDiskReadRecent),
                makeColumn(@"Disk written (recent)", 5 * self.em, NSTextAlignmentRight, @"diskWritten", getDiskWrittenRecent),
                makeColumn(@"Disk read (total)", 5 * self.em, NSTextAlignmentRight, @"totalDiskRead", getDiskReadTotal),
                makeColumn(@"Disk written (total)", 5 * self.em, NSTextAlignmentRight, @"totalDiskWritten", getDiskWrittenTotal),
                        ];
        self.procTable = [[NSTableView alloc] initWithFrame:self.bounds];
        self.procTable.dataSource = self;
        self.procTable.delegate = self;
        self.procTable.intercellSpacing = NSMakeSize(0, 0);
        for (Column *c in self.columns) {
            [self.procTable addTableColumn:c];
        }
        self.procTable.sortDescriptors = @[ ((Column*)[self.columns objectAtIndex:5]).sortDescriptorPrototype ];

        self.procTableContainer = [[NSScrollView alloc] init];
        self.procTableContainer.documentView = self.procTable;
        self.procTableContainer.hasVerticalScroller = YES;
        self.procTableContainer.hasHorizontalScroller = YES;
        [self addSubview:self.procTableContainer];

        self.procTableContainer.autoresizingMask = NSViewWidthSizable | NSViewHeightSizable;

        self.memLabel = createLabel(@"Memory usage");
        self.memLabel.textColor = NSColor.blackColor;
        [self addSubview:self.memLabel];

        self.memBar = [[BarGraph alloc] initWithFrame:CGRectZero];
        self.memBar.frame = NSMakeRect(0, 0, 0, 2 * self.em);
        self.memBar.colors = @[ [NSColor colorWithRed:.75 green:0 blue:0 alpha:1],
                                [NSColor colorWithRed:.75 green:.5 blue:0 alpha:1],
                                [NSColor colorWithRed:.75 green:.75 blue:0 alpha:1],
                                [NSColor colorWithRed:0 green:.75 blue:0 alpha:1],
                                [NSColor colorWithRed:0 green:.5 blue:.5 alpha:1],
                                [NSColor colorWithRed:0 green:0 blue:.75 alpha:1],
                                [NSColor colorWithRed:.5 green:0 blue:.5 alpha:1] ];
        self.memBar.wantsLayer = YES;
        self.memBar.layer.borderWidth = 1.0;
        self.memBar.layer.borderColor = [NSColor grayColor].CGColor;
        [self addSubview:self.memBar];
        
    }
    return self;
}

- (void)layout
{
    [super layout];
    
    NSRect r = self.bounds;
    CGFloat em = self.em;
    CGFloat margin = em;
    CGFloat w = r.size.width - 2 * margin;
    CGFloat x = margin;
    CGFloat y = NSMaxY(r) - margin;

    y -= placeRow(x, y, w, 0, @[ self.memLabel ]);
    self.memBar.frame = NSMakeRect(0, 0, r.size.width - 2 * margin, 2 * self.em);
    y -= placeRow(x, y, w, 0, @[ self.memBar ]);
    y -= em;

    self.procTableContainer.frame = NSMakeRect(x, margin, w, y - margin);
}

- (void)update
{
    NSMutableArray *procs = self.model.processes.mutableCopy;
    [procs sortUsingDescriptors:self.procTable.sortDescriptors];
    self.processes = procs;
    [self.procTable reloadData];

    NSMutableArray *sizes = [[NSMutableArray alloc] initWithCapacity:self.processes.count];
    NSMutableArray *tooltips = [[NSMutableArray alloc] initWithCapacity:self.processes.count];
    double totalMem = 0.0;
    for (Process *p in self.processes) {
        totalMem += p.totalMem;
        [sizes addObject:@(p.totalMem)];
    }
    for (Process *p in self.processes) {
        NSString *name;
        if ([p.localizedProcessName isEqualToString:@""]) {
            name = p.processName;
        } else {
            name = p.localizedProcessName;
        }
        [tooltips addObject:[NSString stringWithFormat:@"%@, %@ (%.1f%%)",
                             name, formattedByteCount((uint64_t)p.totalMem), p.totalMem / totalMem * 100.0]];
    }
    self.memBar.amounts = sizes;
    self.memBar.total = totalMem;
    self.memBar.tooltips = tooltips;
}

// NSTableViewDataSource
- (NSInteger)numberOfRowsInTableView:(NSTableView *)tableView
{
    return self.processes.count;
}

// NSTableViewDelegate
- (CGFloat)tableView:(NSTableView *)tableView heightOfRow:(NSInteger)row
{
    return self.em;
}

- (id)tableView:(NSTableView*)tableView viewForTableColumn:(nullable NSTableColumn *)tableColumn row:(NSInteger)row
{
    Column *c = (Column*)tableColumn;
    Process *p = [self.processes objectAtIndex:row];
    NSTextField *cell = [tableView makeViewWithIdentifier:@"AllProcessesTableView" owner:self];
    if (cell == nil) {
        cell = createLabel(@"");
        cell.identifier = @"AllProcessesTableView";
    }
    cell.textColor = [NSColor blackColor];
    cell.usesSingleLineMode = YES;
    cell.alignment = c.alignment;
    cell.stringValue = c.dataFunc(p);
    if (row % 2 != 0) {
        cell.backgroundColor = [NSColor colorWithRed:.975 green:.975 blue:.975 alpha:1];
    } else {
        cell.backgroundColor = [NSColor whiteColor];
    }
    cell.drawsBackground = YES;
    return cell;
}

- (void)tableView:(NSTableView *)tableView sortDescriptorsDidChange:(NSArray *)oldDescriptors
{
    [self update];
}
@end
