//
//  Preferences.m
//  UpTop
//
//  Created by Geoff Prewett on 11/25/16.
//  Copyright © 2016 Geoff Prewett. All rights reserved.
//

#import "Preferences.h"

#import "FontPicker.h"
#import "NumericEdit.h"
#import "Settings.h"
#import "Utils.h"

#include "math.h"

static NSTextField* addLabel(NSView *parent, NSString *text)
{
    NSTextField* label = [NSTextField labelWithString:text];
    [label sizeToFit];
    [parent addSubview:label];
    return label;
}

static NSTextField* addBoldLabel(NSView *parent, NSString *text)
{
    NSTextField *label = addLabel(parent, text);
    label.font = [NSFontManager.sharedFontManager convertFont:label.font toHaveTrait:NSBoldFontMask];
    [label sizeToFit];
    return label;
}

static NSButton* addButton(NSView *parent, SEL callback, NSString *text)
{
    NSButton *b = [NSButton buttonWithTitle:text target:parent action:callback];
    [b sizeToFit];
    [parent addSubview:b];
    return b;
}

static NSColorWell* addColorWell(NSView *parent, SEL callback, NSColor *color)
{
    NSColorWell *well = [[NSColorWell alloc] initWithFrame:CGRectZero];
    well.color = color;
    well.target = parent;
    well.action = callback;
    CGFloat em = calcEmPx();
    well.frame = NSMakeRect(0, 0, 3 * em, 1.5 * em);
    [parent addSubview:well];
    return well;
}


@interface Preferences ()
@property NSTextField* delayLabel;
@property NumericEdit* delay;
@property NSTextField* delayUnitsLabel;
@property NSTextField* locationLabel;
@property NSPopUpButton* location;
@property NSButton* showLocalizedNames; // has its own label
@property NSTextField* fontLabel;
@property FontPicker* fonts;
@property NumericEdit* fontSize;
@property NSTextField* memoryLabel;
@property NSColorWell* wiredColor;
@property NSTextField* wiredLabel;
@property NSColorWell* compressedColor;
@property NSTextField* compressedLabel;
@property NSColorWell* appColor;
@property NSTextField* appLabel;
@property NSColorWell* swapColor;
@property NSTextField* swapLabel;
@property NSButton* showSwapButton;  // has its own label
@property NSTextField* memoryLinesLabel;
@property NumericEdit* nMemoryLines;
@property NSTextField* cpuLabel;
@property NSColorWell* systemCPUColor;
@property NSTextField* systemCPULabel;
@property NSColorWell* userCPUColor;
@property NSTextField* userCPULabel;
@property NSTextField* cpuLinesLabel;
@property NumericEdit* nCPULines;
@property NSTextField* diskIOLabel;
@property NSTextField* diskIOLinesLabel;
@property NumericEdit* nDiskIOLines;
@property NSTextField* diskIOAccumLabel1;
@property NSTextField* diskIOAccumLabel2;
@property NumericEdit* diskIOAccumLength;
@property NSButton* reset;

@end

@implementation Preferences
- (instancetype)initWithFrame:(NSRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        [self create];
    }
    return self;
}

- (void)create
{
    self.delayLabel = addLabel(self, @"Refresh delay");
    self.delayUnitsLabel = addLabel(self, @"sec");
    
    self.delay = [[NumericEdit alloc] initWithFrame:CGRectZero];
    self.delay.minValue = 0;
    self.delay.maxValue = 3600;
    self.delay.increment = .25;
    self.delay.doubleValue = Settings.refreshDelaySec;
    self.delay.target = self;
    self.delay.action = @selector(onDelayChanged);
    [self.delay sizeToFit];
    self.delay.frame = NSMakeRect(0, 0, 1.1 * self.delay.frame.size.width, self.delay.frame.size.height);
    [self addSubview:self.delay];

    self.locationLabel = addLabel(self, @"Window location");
    self.location = [[NSPopUpButton alloc] initWithFrame:CGRectZero pullsDown:NO];
    [self.location addItemWithTitle:@"Upper right"];
    [self.location addItemWithTitle:@"Middle right"];
    [self.location addItemWithTitle:@"Lower right"];
    [self.location addItemWithTitle:@"Upper left"];
    [self.location addItemWithTitle:@"Middle left"];
    [self.location addItemWithTitle:@"Lower left"];
    [self.location addItemWithTitle:@"Moveable (has titlebar)"];
    [self.location selectItemAtIndex:Settings.windowLocation];
    self.location.target = self;
    self.location.action = @selector(onLocationChanged);
    [self.location sizeToFit];
    [self addSubview:self.location];

    self.showLocalizedNames = [[NSButton alloc] initWithFrame:CGRectZero];
    self.showLocalizedNames.buttonType = NSSwitchButton;
    self.showLocalizedNames.title = @"Show app name instead of executable name";
    self.showLocalizedNames.font = self.delayLabel.font;
    self.showLocalizedNames.state = (Settings.showLocalizedProcessNames ? NSOnState : NSOffState);
    self.showLocalizedNames.target = self;
    self.showLocalizedNames.action = @selector(onShowNamesChanged);
    [self.showLocalizedNames sizeToFit];
    [self addSubview:self.showLocalizedNames];

    self.fontLabel = addBoldLabel(self, @"Font");
    
    self.fonts = [[FontPicker alloc] initWithFrame:CGRectZero];
    self.fonts.target = self;
    self.fonts.action = @selector(onFontChanged);
    [self.fonts sizeToFit];
    [self addSubview:self.fonts];
    
    self.fontSize = [[NumericEdit alloc] initWithFrame:CGRectZero];
    self.fontSize.target = self;
    self.fontSize.action = @selector(onFontChanged);
    [self.fontSize sizeToFit];
    [self addSubview:self.fontSize];
    
    self.memoryLabel = addBoldLabel(self, @"Memory");
    self.memoryLinesLabel = addLabel(self, @"Number of processes to show");
    
    self.wiredLabel = addLabel(self, @"Wired memory (non-swappable kernel memory)");
    self.wiredColor = addColorWell(self, @selector(onColorsChanged), Settings.wiredMemoryColor);
    self.compressedLabel = addLabel(self, @"Compressed memory");
    self.compressedColor = addColorWell(self, @selector(onColorsChanged), Settings.compressedMemoryColor);
    self.appLabel = addLabel(self, @"App memory");
    self.appColor = addColorWell(self, @selector(onColorsChanged), Settings.appMemoryColor);
    self.swapLabel = addLabel(self, @"Swap");
    self.swapColor = addColorWell(self, @selector(onColorsChanged), Settings.swapColor);
    
    self.showSwapButton = [[NSButton alloc] initWithFrame:CGRectZero];
    self.showSwapButton.buttonType = NSSwitchButton;
    self.showSwapButton.title = @"Show swap usage";
    self.showSwapButton.font = self.delayLabel.font;
    self.showSwapButton.state = (Settings.showSwap ? NSOnState : NSOffState);
    self.showSwapButton.target = self;
    self.showSwapButton.action = @selector(onShowSwapChanged);
    [self.showSwapButton sizeToFit];
    [self addSubview:self.showSwapButton];
    
    self.nMemoryLines = [[NumericEdit alloc] initWithFrame:CGRectZero];
    self.nMemoryLines.minValue = 0;
    self.nMemoryLines.maxValue = 20;
    self.nMemoryLines.increment = 1;
    self.nMemoryLines.doubleValue = Settings.nMemoryLines;
    self.nMemoryLines.target = self;
    self.nMemoryLines.action = @selector(onMemoryChanged);
    [self.nMemoryLines sizeToFit];
    [self addSubview:self.nMemoryLines];
    
    self.cpuLabel = addBoldLabel(self, @"CPU Usage");
    self.cpuLinesLabel = addLabel(self, @"Number of processes to show");
    
    self.nCPULines = [[NumericEdit alloc] initWithFrame:CGRectZero];
    self.nCPULines.minValue = 0;
    self.nCPULines.maxValue = 20;
    self.nCPULines.increment = 1;
    self.nCPULines.doubleValue = Settings.nCPULines;
    self.nCPULines.target = self;
    self.nCPULines.action = @selector(onCPUChanged);
    [self.nCPULines sizeToFit];
    [self addSubview:self.nCPULines];

    self.systemCPULabel = addLabel(self, @"System CPU usage");
    self.systemCPUColor = addColorWell(self, @selector(onColorsChanged), Settings.systemCPUColor);
    self.userCPULabel = addLabel(self, @"User CPU usage");
    self.userCPUColor = addColorWell(self, @selector(onColorsChanged), Settings.userCPUColor);

    self.diskIOLabel = addBoldLabel(self, @"Disk I/O");
    self.diskIOLinesLabel = addLabel(self, @"Number of processes to show");
    
    self.nDiskIOLines = [[NumericEdit alloc] initWithFrame:CGRectZero];
    self.nDiskIOLines.minValue = 0;
    self.nDiskIOLines.maxValue = 20;
    self.nDiskIOLines.increment = 1;
    self.nDiskIOLines.doubleValue = Settings.nDiskIOLines;
    self.nDiskIOLines.target = self;
    self.nDiskIOLines.action = @selector(onDiskIOChanged);
    [self.nDiskIOLines sizeToFit];
    [self addSubview:self.nDiskIOLines];

    self.diskIOAccumLabel1 = addLabel(self, @"Accumulate per-process I/O for");
    self.diskIOAccumLabel2 = addLabel(self, @"ticks");
    self.diskIOAccumLength = [[NumericEdit alloc] initWithFrame:CGRectZero];
    self.diskIOAccumLength.minValue = 1;
    self.diskIOAccumLength.maxValue = 20;
    self.diskIOAccumLength.increment = 1;
    self.diskIOAccumLength.doubleValue = Settings.diskIOAccumulatorLength;
    self.diskIOAccumLength.target = self;
    self.diskIOAccumLength.action = @selector(onDiskAccumChanged);
    [self.diskIOAccumLength sizeToFit];
    [self addSubview:self.diskIOAccumLength];
    

    NSFont *f = Settings.font;
    self.fonts.selectedFontName = f.familyName;
    self.fontSize.doubleValue = f.pointSize;
    
    self.reset = addButton(self, @selector(resetDefaults), @"Reset to Default");
}

- (void)sizeToFit
{
    [self layout];
    CGFloat minX = 1e9, maxX = 0, maxY = 0;
    for (NSView *v in self.subviews) {
        minX = MIN(minX, NSMinX(v.frame));
        maxX = MAX(maxX, NSMaxX(v.frame));
        maxY = MAX(maxY, NSMaxY(v.frame));
    }
    CGFloat margin = minX;
    self.frame = NSMakeRect(0, 0, maxX + margin, maxY + margin);
}

- (void)layout
{
    [super layout];

    CGFloat em = self.fontLabel.frame.size.height;
    CGFloat margin = em;
    CGFloat padding = em;
    NSRect r = NSMakeRect(margin, margin, self.frame.size.width - 2 * margin, self.frame.size.height - 2 * margin);
    CGFloat left = r.origin.x;
    CGFloat y = 0;  // placement assumes the window will be resized using -sizeToFit, or that
                    // the window is the correct size

    y -= placeRow(left, y, NSWidth(r), 0, @[ self.delayLabel, self.delay, self.delayUnitsLabel ]);
    y -= placeRow(left, y, NSWidth(r), em, @[ self.locationLabel, self.location ]);
    y -= placeRow(left, y, NSWidth(r), padding, @[ self.showLocalizedNames ]);
    y -= em;

    y = positionXY(self.fontLabel, left, y);
    y -= placeRow(left, y, NSWidth(r), padding, @[ self.fonts, self.fontSize ]);
    y -= em;

    // Memory
    y = positionXY(self.memoryLabel, left, y);
    // Note:  use the height for the label + numeric line, for the checkbox because it is higher;
    //        using the checkbox + label height gives a shorter line and makes it look awkward.
    CGFloat h = MAX(self.nMemoryLines.frame.size.height, self.memoryLinesLabel.frame.size.height);
    position(self.showSwapButton, left, y - (h - self.showSwapButton.frame.size.height) / 2.0,
             self.showSwapButton.frame.size.width, self.showSwapButton.frame.size.height);
    y -= h;
    y -= placeRow(left, y, NSWidth(r), padding, @[ self.memoryLinesLabel, self.nMemoryLines ]);

    y -= placeRow(left, y, NSWidth(r), .25 * em, @[ self.wiredColor, self.wiredLabel ]);
    y -= placeRow(left, y, NSWidth(r), .25 * em, @[ self.compressedColor, self.compressedLabel ]);
    y -= placeRow(left, y, NSWidth(r), .25 * em, @[ self.appColor, self.appLabel ]);
    y -= placeRow(left, y, NSWidth(r), .25 * em, @[ self.swapColor, self.swapLabel ]);

    y -= em;

    // CPU
    y = positionXY(self.cpuLabel, left, y);
    y -= placeRow(left, y, NSWidth(r), padding, @[ self.cpuLinesLabel, self.nCPULines ]);
    y -= placeRow(left, y, NSWidth(r), .25 * em, @[ self.systemCPUColor, self.systemCPULabel ]);
    y -= placeRow(left, y, NSWidth(r), .25 * em, @[ self.userCPUColor, self.userCPULabel ]);
    y -= em;

    // Disk I/O
    y = positionXY(self.diskIOLabel, left, y);
    y -= placeRow(left, y, NSWidth(r), padding, @[ self.diskIOLinesLabel, self.nDiskIOLines ]);
    y -= placeRow(left, y, NSWidth(r), 0, @[ self.diskIOAccumLabel1, self.diskIOAccumLength, self.diskIOAccumLabel2 ]);
    y -= em;

    y -= placeRow(left, y, NSWidth(r), 0, @[ NSNull.null, self.reset, NSNull.null ]);

    // Correct y
    CGFloat minY = 1e9;
    for (NSView *v in self.subviews) {
        minY = MIN(minY, v.frame.origin.y);
    }
    minY -= margin;
    CGFloat dy = fabs(minY);
    for (NSView *v in self.subviews) {
        v.frame = NSMakeRect(v.frame.origin.x, v.frame.origin.y + dy, v.frame.size.width, v.frame.size.height);
    }
}

- (void)onDelayChanged
{
    [Settings setDefaultRefreshDelay:self.delay.doubleValue];
    [NSNotificationCenter.defaultCenter postNotificationName:@"DelayChanged" object:self];
}

- (void)onLocationChanged
{
    Location newLoc = (Location)self.location.indexOfSelectedItem;
    // You'd think this would be a good place to save the window frame, but for some
    // reason self.window.frame is incorrect (at least compared to the values in
    // the app delegate, where it was created...)
    [Settings setDefaultPosition:newLoc];
    [self sendUpdateUI];
}

- (void)onShowNamesChanged
{
    [Settings setShowLocalizedProcessNames:(Settings.showLocalizedProcessNames == YES ? NO : YES)];
    [self sendUpdateUI];
}

- (void)onFontChanged
{
    [Settings setDefaultFontName:self.fonts.selectedFontName size:self.fontSize.doubleValue];
    [self sendUpdateUI];
}

- (void)onMemoryChanged
{
    [Settings setDefaultNMemoryLines:(int)self.nMemoryLines.doubleValue];
    [self sendUpdateUI];
}

- (void)onCPUChanged
{
    [Settings setDefaultNCPULines:(int)self.nCPULines.doubleValue];
    [self sendUpdateUI];
}

- (void)onDiskIOChanged
{
    [Settings setDefaultNDiskIOLines:(int)self.nDiskIOLines.doubleValue];
    [self sendUpdateUI];
}

- (void)onShowSwapChanged
{
    [Settings setShowSwap:(self.showSwapButton.state == NSOnState)];
    [self sendUpdateUI];
}

- (void)onDiskAccumChanged
{
    [Settings setDefaultDiskIOAccumulatorLength:self.diskIOAccumLength.doubleValue];
    [NSNotificationCenter.defaultCenter postNotificationName:@"RecreateModel" object:self];
}

- (void)onColorsChanged
{
    [Settings setDefaultMemoryColors:self.wiredColor.color compressed:self.compressedColor.color
                                 app:self.appColor.color swap:self.swapColor.color];
    [Settings setDefaultCPUColors:self.systemCPUColor.color user:self.userCPUColor.color];
    [self sendUpdateUI];
}

- (void)resetDefaults
{
    [Settings resetToDefault];
    while (self.subviews.count > 0) {
        [[self.subviews objectAtIndex:0] removeFromSuperviewWithoutNeedingDisplay];
    }
    [self create];
    [self layout];
    self.needsDisplay = YES;
    [self sendUpdateUI];
}

- (void)sendUpdateUI
{
    [NSNotificationCenter.defaultCenter postNotificationName:@"UpdateUI" object:self];
}

@end
