//
//  ProcessTable.m
//  UpTop
//
//  Created by Geoff Prewett on 11/14/16.
//  Copyright © 2016 Geoff Prewett. All rights reserved.
//

#import "ProcessTable.h"

#import "Settings.h"
#import "Utils.h"

#define kMB     1048576.0

@interface ProcessTable ()
@property NSArray<Process*>* data;
@property int nCellsCreated;

@property enum ProcessStat processStat;
@property NSTableColumn* colOne;
@property NSTableColumn* colTwo;
@property NSTableColumn* colThree;
@property NSTableView* table;
@property NSScrollView* tableContainer;
@end

@implementation ProcessTable
- (instancetype)initFor:(enum ProcessStat)processStat;
{
    self = [super initWithFrame:CGRectZero];
    if (self) {
        NSTextField *tf = createLabel(@"Ag");
        [tf sizeToFit];
        // Hack:  not sure why I need +2, but it clips off descenders otherwise.
        self->_cellHeight = ceil(tf.frame.size.height) + 2;

        self.processStat = processStat;
        self.data = @[];
        self.table = [[NSTableView alloc] initWithFrame:self.bounds];
        self.table.dataSource = self;
        self.table.delegate = self;

        self.colOne = [[NSTableColumn alloc] initWithIdentifier:@"Col1"];
        self.colOne.title = @"Name";
        self.colOne.resizingMask = NSTableColumnNoResizing;
        [self.table addTableColumn:self.colOne];
        self.colTwo = [[NSTableColumn alloc] initWithIdentifier:@"Col2"];
        self.colTwo.title = @"Value";
        self.colTwo.width = 4 * self.cellHeight;
        self.colTwo.resizingMask = NSTableColumnNoResizing;  // .width is only effective if not resizing
        [self.table addTableColumn:self.colTwo];
        if (processStat == kProcDiskIO) {
            self.colTwo.width = 3 * self.cellHeight;
            self.colThree = [[NSTableColumn alloc] initWithIdentifier:@"Col3"];
            self.colThree.title = @"Value2";
            self.colThree.width = self.colTwo.width;
            self.colThree.resizingMask = NSTableColumnNoResizing;
            [self.table addTableColumn:self.colThree];
        }

        self.table.backgroundColor = [NSColor clearColor];
        self.table.headerView = nil;
        self.table.intercellSpacing = NSMakeSize(0, 0);

        self.tableContainer = [[NSScrollView alloc] init];
        self.tableContainer.documentView = self.table;
        self.tableContainer.hasVerticalScroller = YES;
        [self addSubview:self.tableContainer];

        self.tableContainer.autoresizingMask = NSViewWidthSizable | NSViewHeightSizable;
        self.tableContainer.drawsBackground = NO;
    }
    return self;
}

- (void)setFrame:(NSRect)frame
{
    [super setFrame:frame];
    // NSTableViewFirstColumnOnlyAutoresizingStyle seems to feel free to set the table width,
    // and I can't figure out any way to prevent it, so do it manually.
    CGFloat colOneWidth = self.bounds.size.width - self.colTwo.width;
    if (self.colThree != nil) {
        colOneWidth -= self.colThree.width;
    }
    self.colOne.width = colOneWidth;
}

- (void)setProcesses:(NSArray<Process*>*)procs
{
//    self.data = procs;
//    [self.table reloadData];
    if (procs != nil && self.data != nil && self.data.count == procs.count) {
        self.data = procs;
        NSMutableIndexSet *rows = [[NSMutableIndexSet alloc] init];
        [rows addIndexesInRange:NSMakeRange(0, procs.count)];
        NSMutableIndexSet *cols = [[NSMutableIndexSet alloc] init];
        [cols addIndexesInRange:NSMakeRange(0, self.table.tableColumns.count)];
        [self.table reloadDataForRowIndexes:rows columnIndexes:cols];
    } else {
        self.data = procs;
        [self.table reloadData];
    }
}

// NSTableViewDataSource
- (NSInteger)numberOfRowsInTableView:(NSTableView *)tableView
{
    return self.data.count;
}

// NSTableViewDelegate
- (CGFloat)tableView:(NSTableView *)tableView heightOfRow:(NSInteger)row
{
    return self.cellHeight;
}

- (BOOL)selectionShouldChangeInTableView:(NSTableView *)tableView
{
    return NO;
}

- (id)tableView:(NSTableView*)tableView viewForTableColumn:(nullable NSTableColumn *)tableColumn row:(NSInteger)row
{
    NSString *value = @"nil";
    Process *p = [self.data objectAtIndex:row];
    if (tableColumn == self.colOne) {
        if (Settings.showLocalizedProcessNames && ![p.localizedProcessName isEqualToString:@""]) {
            value = p.localizedProcessName;
        } else {
            value = p.processName;
        }
    } else if (tableColumn == self.colTwo) {
        if (self.processStat == kProcMemory) {
            value = [NSString stringWithFormat:@"%.1f MB", p.totalMem / kMB];
        } else if (self.processStat == kProcCPU) {
            value = [NSString stringWithFormat:@"%.1f%%", p.cpuPercent];
        } else if (self.processStat == kProcDiskIO) {
            if (p.diskRead == 0) {
                value = @"";
            } else {
                value = formattedByteCount(p.diskRead);
            }
        }
    } else if (tableColumn == self.colThree) {
        if (self.processStat == kProcDiskIO) {
            if (p.diskWritten == 0) {
                value = @"";
            } else {
                value = formattedByteCount(p.diskWritten);
            }
        }
    }

    NSTextField *cell = [tableView makeViewWithIdentifier:@"ProcessTableView" owner:self];
    if (cell == nil) {
        cell = createLabel(@"");
        cell.identifier = @"ProcessTableView";
    }
/*    cell.wantsLayer = YES;
    cell.layer.borderWidth = 1.0;
    cell.layer.borderColor = [NSColor grayColor].CGColor; */
    if (p.notCurrent == YES) {
        cell.textColor = [NSColor grayColor];
    } else {
        cell.textColor = Settings.foregroundColor;
    }
    cell.usesSingleLineMode = YES;
    if (tableColumn == self.colOne) {
        cell.alignment = NSTextAlignmentLeft;
    } else {
        cell.alignment = NSTextAlignmentRight;
    }
    cell.stringValue = value;
//    cell.toolTip = [NSString stringWithFormat:@"binary: %@\nname: %@", p.processName, p.localizedProcessName];
    return cell;
}
@end
