//
//  Settings.m
//  UpTop
//
//  Created by Geoff Prewett on 11/14/16.
//  Copyright © 2016 Geoff Prewett. All rights reserved.
//

#import "Settings.h"

#define kRefreshDelayKey                @"RefreshDelaySec"
#define kLocationKey                    @"WindowLocation"
#define kWindowRectKey                  @"WindowRect"
#define kFontNameKey                    @"FontName"
#define kFontSizeKey                    @"FontSize"
#define kWiredMemoryColorKey            @"WiredMemoryColor"
#define kCompressedMemoryColorKey       @"CompressedMemoryColor"
#define kAppMemoryColorKey              @"AppMemoryColor"
#define kSwapColorKey                   @"SwapColor"
#define kSystemCPUColorKey              @"SystemCPUColor"
#define kUserCPUColorKey                @"UserCPUColor"
#define kMemoryNLinesKey                @"MemoryNLines"
#define kCPUNLinesKey                   @"CPUNLines"
#define kDiskIONLinesKey                @"DiskIONLines"
#define kShowSwapKey                    @"ShowSwap"
#define kDiskIOAccumulatorLengthKey     @"DiskIOAccumulatorLength"
#define kShowLocalizedProcessNamesKey   @"ShowLocalizedProcessNames"

#define kDefaultRefreshDelaySec         1.0
#define kDefaultLocation                kLocUpperRight
#define kDefaultWindowRect              CGRectZero
#define kDefaultNLines                  3
#define kDefaultShowSwap                YES
#define kDefaultShowLocalizedProcessNames YES
#define kDefaultDiskIOAccumulatorLength 5
#define kDefaultWiredMemoryColor        [NSColor colorWithRed:.5 green:.5 blue:.5  alpha:1]
#define kDefaultCompressedMemoryColor   [NSColor colorWithRed:.4 green:0  blue:.75 alpha:1]
#define kDefaultAppMemoryColor          [NSColor colorWithRed:0  green:0  blue:.75 alpha:1]
#define kDefaultSwapColor               [NSColor colorWithRed:.6 green:0 blue:0  alpha:1]
#define kDefaultSystemCPUColor          [NSColor colorWithRed:.6 green:0  blue:0 alpha:1]
#define kDefaultUserCPUColor            [NSColor colorWithRed:.6 green:.3 blue:0 alpha:1]

static NSFont *gCachedFont = nil;

static int getInt(NSString *key, int defaultValue)
{
    NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
    if ([defaults objectForKey:key] == nil) {
        return defaultValue;
    }
    int n = (int)[defaults integerForKey:key];
    return n;
}

static NSColor* getColor(NSString *key, NSColor *defaultColor)
{
    NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
    if ([defaults objectForKey:key] == nil) {
        return defaultColor;
    }
    NSData *data = [defaults dataForKey:key];
    if (!data) {
        return defaultColor;
    }
    return (NSColor*)[NSUnarchiver unarchiveObjectWithData:data];
}

static void setColor(NSString *key, NSColor *color)
{
    NSData *data = [NSArchiver archivedDataWithRootObject:color];  // NSColor cannot be used in user defaults
    NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
    [defaults setObject:data forKey:key];
}

@implementation Settings : NSObject
+ (double)refreshDelaySec
{
    NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
    if ([defaults objectForKey:kRefreshDelayKey] == nil) {
        return kDefaultRefreshDelaySec;
    }
    double delaySec = (double)[defaults doubleForKey:kRefreshDelayKey];
    return delaySec;
}

+ (int)diskIOAccumulatorLength
{
    return getInt(kDiskIOAccumulatorLengthKey, kDefaultDiskIOAccumulatorLength);
}

+ (Location)windowLocation
{
    return getInt(kLocationKey, kDefaultLocation);
}

+ (NSRect)windowRect
{
    NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
    if ([defaults objectForKey:kWindowRectKey] == nil) {
        return kDefaultWindowRect;
    }
    NSString *rectStr = [defaults valueForKey:kWindowRectKey];
    if (!rectStr) {
        return kDefaultWindowRect;
    }
    return NSRectFromString(rectStr);
}

+ (NSColor*)foregroundColor
{
    return [NSColor whiteColor];
}

+ (NSColor*)wiredMemoryColor
{
    return getColor(kWiredMemoryColorKey, kDefaultWiredMemoryColor);
}

+ (NSColor*)compressedMemoryColor
{
    return getColor(kCompressedMemoryColorKey, kDefaultCompressedMemoryColor);
}

+ (NSColor*)appMemoryColor
{
    return getColor(kAppMemoryColorKey, kDefaultAppMemoryColor);
}

+ (NSColor*)swapColor
{
    return getColor(kSwapColorKey, kDefaultSwapColor);
}

+ (NSColor*)systemCPUColor
{
    return getColor(kSystemCPUColorKey, kDefaultSystemCPUColor);
}

+ (NSColor*)userCPUColor
{
    return getColor(kUserCPUColorKey, kDefaultUserCPUColor);
}

+ (NSFont*)font
{
    if (gCachedFont == nil) {
        NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
        NSString *fontName = [defaults stringForKey:kFontNameKey];
        double fontSize = [defaults doubleForKey:kFontSizeKey];
        if (fontName == nil && fontSize == 0) {
            gCachedFont = [NSFont labelFontOfSize:12];
        } else if (fontName == nil) {
            gCachedFont = [NSFont labelFontOfSize:fontSize];
        } else {
            gCachedFont = [NSFont fontWithName:fontName size:fontSize];
        }
        // Just in case something failed...
        if (gCachedFont == nil) {
            gCachedFont = [NSFont labelFontOfSize:12];
        }
    }
    return gCachedFont;
}

+ (int)nMemoryLines
{
    return getInt(kMemoryNLinesKey, kDefaultNLines);
}

+ (int)nCPULines
{
    return getInt(kCPUNLinesKey, kDefaultNLines);
}

+ (int)nDiskIOLines
{
    return getInt(kDiskIONLinesKey, kDefaultNLines);
}

+ (BOOL)showSwap
{
    return (BOOL)getInt(kShowSwapKey, kDefaultShowSwap);
}

+ (BOOL)showLocalizedProcessNames
{
    return (BOOL)getInt(kShowLocalizedProcessNamesKey, kDefaultShowLocalizedProcessNames);
}

+ (void)setDefaultRefreshDelay:(double)sec
{
    NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
    [defaults setDouble:sec forKey:kRefreshDelayKey];
    [defaults synchronize];
}

+ (void)setDefaultPosition:(Location)loc
{
    NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
    [defaults setInteger:loc forKey:kLocationKey];
    [defaults synchronize];
}

+ (void)setDefaultWindowRect:(NSRect)r
{
    NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
    [defaults setValue:NSStringFromRect(r) forKey:kWindowRectKey];
    [defaults synchronize];
}

+ (void)setDefaultMemoryColors:(NSColor*)wired compressed:(NSColor*)compressed app:(NSColor*)app swap:(NSColor*)swap
{
    setColor(kWiredMemoryColorKey, wired);
    setColor(kCompressedMemoryColorKey, compressed);
    setColor(kAppMemoryColorKey, app);
    setColor(kSwapColorKey, swap);
}

+ (void)setDefaultCPUColors:(NSColor*)system user:(NSColor*)user
{
    setColor(kSystemCPUColorKey, system);
    setColor(kUserCPUColorKey, user);
}

+ (void)setDefaultFontName:(NSString*)fontName size:(CGFloat)fontSize
{
    NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
    [defaults setObject:fontName forKey:kFontNameKey];
    [defaults setDouble:fontSize forKey:kFontSizeKey];
    [defaults synchronize];
    gCachedFont = nil;
}

+ (void)setDefaultNMemoryLines:(int)nLines
{
    NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
    [defaults setInteger:nLines forKey:kMemoryNLinesKey];
    [defaults synchronize];
}

+ (void)setDefaultNCPULines:(int)nLines
{
    NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
    [defaults setInteger:nLines forKey:kCPUNLinesKey];
    [defaults synchronize];
}

+ (void)setDefaultNDiskIOLines:(int)nLines
{
    NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
    [defaults setInteger:nLines forKey:kDiskIONLinesKey];
    [defaults synchronize];
}

+ (void)setShowSwap:(BOOL)show
{
    NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
    [defaults setInteger:show forKey:kShowSwapKey];
    [defaults synchronize];
}

+ (void)setDefaultDiskIOAccumulatorLength:(int)n
{
    NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
    [defaults setInteger:n forKey:kDiskIOAccumulatorLengthKey];
    [defaults synchronize];
}

+ (void)setShowLocalizedProcessNames:(BOOL)show
{
    NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
    [defaults setInteger:show forKey:kShowLocalizedProcessNamesKey];
    [defaults synchronize];
}

+ (void)resetToDefault
{
    // You'd think +resetStandardUserDefaults would be useful here, but it doesn't
    // seem to do anything.

    NSUserDefaults *defaults = NSUserDefaults.standardUserDefaults;
    for (NSString *key in defaults.dictionaryRepresentation) {
        [defaults removeObjectForKey:key];
    }
    [defaults synchronize];

    gCachedFont = nil;
}

@end
