//
//  TopView.m
//  UpTop
//
//  Created by Geoff Prewett on 11/13/16.
//  Copyright © 2016 Geoff Prewett. All rights reserved.
//

#import "TopView.h"

#import <CoreImage/CoreImage.h>

#import "BarGraph.h"
#import "ProcessTable.h"
#import "Settings.h"
#import "Utils.h"

static NSTextField* addLabel(NSView *parent, NSString *text)
{
    NSTextField *label = createLabel(text);
    [label sizeToFit];
    [parent addSubview:label];
    return label;
}

@interface TopView ()
@property NSTextField* errorLabel;
@property NSTextField* memoryLabel;
@property BarGraph* memoryBar;
@property NSTextField* freeMemory;
@property BarGraph* swapBar;
@property NSTextField* swapUsedLabel;
@property ProcessTable* procsMemTable;
@property NSTextField* cpuLabel;
@property NSTextField* cpuAmount;
@property BarGraph* cpuBar;
@property ProcessTable* procsCPUTable;
@property NSTextField* diskLabel;
@property NSTextField* diskReadLabel;
@property NSTextField* diskWrittenLabel;
@property ProcessTable* procsDiskTable;
@end

@implementation TopView
- (id)initWithFrame:(NSRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        self.model = nil;

        self.wantsLayer = YES;
        self.layer.backgroundColor = [NSColor colorWithRed:0 green:0 blue:0 alpha:.75].CGColor;
        if (Settings.windowLocation < kLocMoveable) {
            self.layer.borderWidth = 1.0;
            self.layer.borderColor = [NSColor grayColor].CGColor;
        }

/*        CIFilter *blur = [CIFilter filterWithName:@"CIGaussianBlur"];
        [blur setDefaults];
//        self.layer.backgroundFilters = [NSArray arrayWithObject: blur];
        self.layer.compositingFilter = blur;
        self.layerUsesCoreImageFilters = YES; */

        self.memoryLabel = addLabel(self, @"Memory");

        self.memoryBar = [[BarGraph alloc] initWithFrame:CGRectZero];
        self.memoryBar.colors = @[ Settings.wiredMemoryColor,
                                   Settings.compressedMemoryColor,
                                   Settings.appMemoryColor ];
        self.memoryBar.wantsLayer = YES;
        self.memoryBar.layer.borderWidth = 1.0;
        self.memoryBar.layer.borderColor = [NSColor grayColor].CGColor;
        [self addSubview:self.memoryBar];

        self.freeMemory = addLabel(self, @"Ag");

        self.swapBar = [[BarGraph alloc] initWithFrame:CGRectZero];
        self.swapBar.colors = @[ Settings.swapColor ];
        self.swapBar.wantsLayer = YES;
        self.swapBar.layer.borderWidth = 1.0;
        self.swapBar.layer.borderColor = [NSColor grayColor].CGColor;
        [self addSubview:self.swapBar];

        self.swapUsedLabel = addLabel(self, @"Ag");

        self.procsMemTable = [[ProcessTable alloc] initFor:kProcMemory];
        [self addSubview:self.procsMemTable];

        self.cpuLabel = addLabel(self, @"CPU Usage");
        self.cpuAmount = addLabel(self, @"Ag");

        self.cpuBar = [[BarGraph alloc] initWithFrame:CGRectZero];
        self.cpuBar.total = 100.0;
        self.cpuBar.colors = @[ Settings.systemCPUColor, Settings.userCPUColor ];
        self.cpuBar.wantsLayer = YES;
        self.cpuBar.layer.borderWidth = 1.0;
        self.cpuBar.layer.borderColor = [NSColor grayColor].CGColor;
        [self addSubview:self.cpuBar];

        self.procsCPUTable = [[ProcessTable alloc] initFor:kProcCPU];
        [self addSubview:self.procsCPUTable];

        self.diskLabel = addLabel(self, @"Disk I/O");
        self.diskReadLabel = addLabel(self, @"Ag");
        self.diskWrittenLabel = addLabel(self, @"Ag");
        self.diskWrittenLabel.alignment = NSTextAlignmentRight;

        self.procsDiskTable = [[ProcessTable alloc] initFor:kProcDiskIO];
        [self addSubview:self.procsDiskTable];

        // Add error last, so it will go on top of everything
        self.errorLabel = addLabel(self, @"");
        self.errorLabel.textColor = [NSColor redColor];
        self.errorLabel.hidden = YES;

        [self update];
    }
    return self;
}

- (void)setError:(NSString*)err
{
    if (self.errorLabel.hidden == YES) {
        self.errorLabel.stringValue = err;
        self.errorLabel.hidden = NO;
        [self.errorLabel sizeToFit];
    }
}

- (void)sizeToFit
{
    self.frame = NSMakeRect(0, 0, 200, 300);
    [self layout];
    self.frame = NSMakeRect(0, 0, self.frame.size.width, ceil(self.frame.size.height - NSMinY(self.procsDiskTable.frame)));
    // No need to layout again, since the positions won't change
}

- (void)layout
{
    [super layout];

    CGFloat em = self.memoryLabel.frame.size.height;
    CGFloat margin = 0.33 * em;
    NSRect r = NSMakeRect(margin, margin, self.frame.size.width - 2 * margin, self.frame.size.height - 2 * margin);
    CGFloat y = self.frame.size.height - margin;
    self.memoryLabel.frame = NSMakeRect(NSMinX(r), y - em,
                                        self.memoryLabel.frame.size.width, self.memoryLabel.frame.size.height);
    y -= self.memoryLabel.frame.size.height;

    CGFloat w = 4.55 * em;
    self.freeMemory.frame = NSMakeRect(NSMaxX(r) - w, y - em, w, self.freeMemory.frame.size.height);
    self.memoryBar.frame = NSMakeRect(NSMinX(r), y - em, NSMinX(self.freeMemory.frame) - NSMinX(r), em);
    y -= em;

    w = NSWidth(r) / 2.0;
    CGFloat h = (Settings.showSwap == YES ? em : 0);
    self.swapBar.frame = NSMakeRect(NSMinX(r), y - em, w, h);
    self.swapUsedLabel.frame = NSMakeRect(NSMinX(r) + w, y - h, w, h);
    y -= h;

    CGFloat tableHeight = Settings.nMemoryLines * self.procsMemTable.cellHeight;
    self.procsMemTable.frame = NSMakeRect(NSMinX(r), y - tableHeight, NSMaxX(r) - NSMinX(r), tableHeight);
    y -= tableHeight;
    y -= em;

    self.cpuLabel.frame = NSMakeRect(NSMinX(r), y - em,
                                     self.cpuLabel.frame.size.width, self.cpuLabel.frame.size.height);
    y -= self.cpuLabel.frame.size.height;

//    w = 3 * em;
    self.cpuAmount.frame = NSMakeRect(NSMaxX(r) - w, y - em, w, self.cpuAmount.frame.size.height);
    self.cpuBar.frame = NSMakeRect(NSMinX(r), y - em, NSMinX(self.cpuAmount.frame) - NSMinX(r), em);
    y -= em;

    tableHeight = Settings.nCPULines * self.procsMemTable.cellHeight;
    self.procsCPUTable.frame = NSMakeRect(NSMinX(r), y - tableHeight, NSMaxX(r) - NSMinX(r), tableHeight);
    y -= tableHeight;
    y -= em;

    self.diskLabel.frame = NSMakeRect(NSMinX(r), y - em,
                                      self.diskLabel.frame.size.width, self.diskLabel.frame.size.height);
    y -= self.diskLabel.frame.size.height;
    CGFloat half = (NSMaxX(r) - NSMinX(r)) / 2.0;
    self.diskReadLabel.frame = NSMakeRect(NSMinX(r), y - em, half, self.diskReadLabel.frame.size.height);
    self.diskWrittenLabel.frame = NSMakeRect(NSMinX(r) + half, y - em, half, self.diskWrittenLabel.frame.size.height);
    y -= em;

    tableHeight = Settings.nDiskIOLines * self.procsMemTable.cellHeight;
    self.procsDiskTable.frame = NSMakeRect(NSMinX(r), y - tableHeight, NSMaxX(r) - NSMinX(r), tableHeight);
    y -= tableHeight;

    self.errorLabel.frame = NSMakeRect(NSMinX(r), NSMaxY(r) - em, NSWidth(r), em);
}

- (void)update
{
    if (self.model.error != nil) {
        [self setError:self.model.error];
        return;
    }

    if (self.model.isCPUvalid) {
        self.cpuBar.amounts = @[ @(self.model.cpuSystemPercentage),
                                 @(self.model.cpuUserPercentage) ];
        self.cpuAmount.stringValue = [NSString stringWithFormat:@"%.1f%%",
                                      self.model.cpuSystemPercentage + self.model.cpuUserPercentage];
    } else {
        self.cpuBar.amounts = @[ @(1), @(1) ];
        self.cpuAmount.stringValue = @"--";
    }

    double gb = 1024.0 * 1024.0 * 1024.0;
    self.memoryBar.total = self.model.memoryTotal;
    self.memoryBar.amounts = @[ @(self.model.memoryWired), @(self.model.memoryCompressed), @(self.model.memoryApps) ];
    self.freeMemory.stringValue = [NSString stringWithFormat:@"%.1f GB avail",
                                   (self.model.memoryTotal - self.model.memoryUsed) / gb];
    self.swapBar.total = self.model.swapTotal;
    self.swapBar.amounts = @[ @(self.model.swapUsed) ];
    self.swapUsedLabel.stringValue = [NSString stringWithFormat:@"%.1f GB swapped", (self.model.swapUsed / gb)];

    self.diskReadLabel.stringValue = [NSString stringWithFormat:@"Read: %@", formattedByteCount(self.model.diskBytesRead)];
    self.diskWrittenLabel.stringValue = [NSString stringWithFormat:@"Written: %@",
                                         formattedByteCount(self.model.diskBytesWritten)];

    NSArray *procs = self.model.processes;
    NSSortDescriptor *sort = [[NSSortDescriptor alloc] initWithKey:@"totalMem" ascending:NO];
    NSArray *memSortedProcs = [procs sortedArrayUsingDescriptors:@[sort]];
    [self.procsMemTable setProcesses:memSortedProcs];

    sort = [[NSSortDescriptor alloc] initWithKey:@"cpuPercent" ascending:NO];
    NSArray *cpuSortedProcs = [procs sortedArrayUsingDescriptors:@[sort]];
    [self.procsCPUTable setProcesses:cpuSortedProcs];

    [self.procsDiskTable setProcesses:self.model.diskPocesses];
}

@end
